/*
 * CCube.cpp
 *
 *  Created on: Dec 5, 2013
 *      Author: Jono
 */




#include "CApproximation.h"


/*
***************************
******   PUBLIC     *******
***************************
*/

//  Generic Constructor
CApproximation::CApproximation(){
	Define_Maximum_Depth(7);
	Define_Interval_Subdivisions(3);
}

/* Copy constructor */
CApproximation::CApproximation(const CApproximation& approx){
	Maximum_Depth = approx.Maximum_Depth;
	poly_ = approx.poly_;
}

// Destructor
CApproximation::~CApproximation(){
}




/*
It is more or less assumed we are dealing with two nodal approximations corresponding to thresholds a<b.

We modify the complex B, so that |B| \subseteq |A|
*/
void CApproximation::Intersect( CComplex & complex_A, CComplex & complex_B){
  /* We go through each cube in the Grid B */
  /* Initialize the stack of CNode pointers we must iterate through */
  std::stack< CNode* > node_stack;
  
  /* We obtain a pointer to the root of complex_B */
  std::vector< unsigned int > blank ;
  CNode * pt_root = complex_B.get_node( blank);
  /* We add the pointer to root to the stack */
  node_stack.push(pt_root);
  while( !(node_stack.empty())){
    /* We remove the top CNode from the stack */
    CNode * current_node = node_stack.top();
    node_stack.pop();
    /* we check if the current node is a leaf */
    if( (*current_node).leaf_check()){
      /* If the current node is a leaf, then we iterate through all of its faces */
      for( unsigned int i = 0 ; i < 9 ; i++){
        if((*current_node).Check_Cell(i)){
          /* We have a cell \lambda in the complex with face_number_ = i */
          SFace lambda = { complex_B.get_location( current_node ) , current_node , i } ;
          /* We need to check if there exists on unique cell \mu in complex_A which contains \lambda */
          /* Here unique referes to the geometric representation. there may very well be multiple computer representations*/
          std::vector < SFace > cover = complex_A.Obtain_Carrier(lambda);
          /* We check to see if our cover for lambda is in complex A. */
          bool lambda_intersects_complex_A = false;
          for( int j = 0 ; j < cover.size(); j++){
            if( cover[j].Check_Cell())
              lambda_intersects_complex_A = true;
          }
          /* If lambda does not intersect complex A, then we delete it */
          if( !(lambda_intersects_complex_A)){
            lambda.Delete();
          }
        }// in complex
      }// lambda
    }// node is leaf
    else{
      /*If current node is not a leaf, then we add all of its children to the stack */ 
      for( unsigned int i =0; i < 4 ;i++){
        node_stack.push( (*current_node).get_child( i ) );
      }
    }
  }
}






/*
For a complex passed by reference, we create a verified approximation corresponding to the given threshold .
*/
bool CApproximation::Verified_Approximation(double threshold, bool sublevel, CComplex & complex){
  /* We initialize a counter to see how many total cubes are tested */
  int total_cubes = 0; 
  /* We initialize a counter to see how many cubes were verified */
  int verified_cube_counter =0;
  /* We initialize a counter to see the maximum depth of the grid */
  int depth_counter =0;
  
  /* We define a boolean variable to determine whether we were successful */ 
  bool success = true;
  /* Initializes the domain of our function */
  CCube unit_cube;
  /* Initialize the stack of cubes we must verify */
  std::stack< CCube > Cubes_to_Check;
  /* Add the unit cube to the stack of cubes to verify */
  Cubes_to_Check.push(unit_cube);
  /* Keep trying to verify cubes as long as there are cubes in the stack */
  while( !(Cubes_to_Check.empty() ) ) {
    /* Remove the first cube from the stack */
    CCube Current_Cube = Cubes_to_Check.top();
    /* Remove the first cube from the stack */
    Cubes_to_Check.pop();
    
    /* We increase the total cube counter */
    total_cubes++;
    /* We update the maximum depth  */
    std::vector< unsigned int > location = Current_Cube.get_location();
    if( location.size() > depth_counter){
      depth_counter =location.size();
    }
    
    /* Apply the Verify algorithm to the current cube */
    int Result = Current_Cube.verify(poly_,threshold );
    /* Now decide on a course of action */
    if( Result == 1){
        /* The cube was verified */
		/* We determine which verticies of the cube are above the threshold */
		std::vector < bool > cube_verticies = Current_Cube.local_approximation(poly_, threshold);
		if (sublevel){
			/* If we are constructing a sublevel set, we need to flip the sign of all the verticies */
			for (int i = 0; i < cube_verticies.size(); i++){
				cube_verticies[i] = !(cube_verticies[i]);
			}
		}
		/* If sublevel is false, then we don't need to modify cube_verticies */
		complex.Add_Cube( Current_Cube.get_location() , cube_verticies );
		/* We add one to the verified cube counter */
		verified_cube_counter ++;
    }
    else if( Result == -1){
      /*Either the maximum depth was reached, or a vertex evaluated to zero */
      std::cout << "Failed to verify threshold " << threshold << "  ::  A vertex interesects the level-set.\n";
      success = false;
      break;
    }
    else if( Current_Cube.get_location().size() > Maximum_Depth){
		/* The maximum depth was reached, or a vertex evaluated to zero */
		std::cout <<"Failed to verify threshold " << threshold << "  :: Exceeded maximum depth of " << Maximum_Depth << ".\n";
		/*This following line of code will print the geometric location of the last cube which failed to be verified. */
		//Current_Cube.GetGeometricRepresentation().Print();
		success = false;
      break;
    }
    else{
      /* The cube was not verified. We subdivide the cube. */
      std::vector< CCube > Subdivided_Cubes = Current_Cube.Subdivide();
      /* Add the resultant cubes to the stack of cubes to check */
      for(int i =0 ; i< Subdivided_Cubes.size()  ; i++){
        Cubes_to_Check.push(Subdivided_Cubes[i]);
      }
    }
  } 
  /* We print out how many cubes were verified */
   //std::cout <<"A total of " << verified_cube_counter << " cubes were verified. \n";
  if( success){
    std::cout << "Verified the threshold " << threshold << "\n";
  }
  
  return success;
}


/*
Only two complexes are stored at once.
If the first threshold does not get verified, this function will exit.

If creating a sublevel set filtration (sublevel == true), then you start at the top threshold and go down.
If creating a superlevel set filtration (sublevel == false), then you start at the bottom threshold and go up.

step_size SHOULD BE POSITIVE
*/
void CApproximation::Verified_Filtration(double bottom_threshold, double step_size, unsigned int max_steps, bool sublevel, CFiltration & filter){
	/* We calculate what the top threshold is */
	double top_threshold = bottom_threshold + step_size * (max_steps - 1);
	/* We initialize the double variable which stores the thresholds we compute verified approximations at */
	double local_threshold;
	// Before we can compute the persistent homology of the function,
    // it is necessary to initialize the filtration with the total number thresholds in our filtration.
  filter.Initialize(max_steps);  
  /* We initialize one pointers to a complexe, and allocate some memory for it*/
  CComplex * pt_complex_A = new CComplex;

  /* We create  the filtration by only storing two complexes at a time */
  for(int i = 0; i < max_steps; i ++ ){
    /* We determine the new threshold, depending on sub/superlevel sets */ 
	if (sublevel){
	  /* Sub-level case*/
	  local_threshold = top_threshold    - i * step_size;
	}
	else{
	  /* Super-level case*/
	  local_threshold = bottom_threshold + i * step_size;
	}
    /* We initialize one pointer-to-complex, and allocate some memory for it*/
    CComplex * pt_complex_B = new CComplex;
    /* Construct the nodal approximation at the local threshold */
    bool success = Verified_Approximation( local_threshold ,sublevel , (* pt_complex_B ) );
    /* If we successfully verified the function at the local_threshold , we add it to the filtration */
    if(success ){
      /*  We need to remove any redundant cells */
      (* pt_complex_B ).Create_CW_Complex();
      /* If this is not the first complex we've verified, we need to intersect the complexes */
      if( i > 0 ){
        Intersect( (*pt_complex_A) , (*pt_complex_B) );
      }
      /* We now add the complex to the filtration */
	  /* NOTE we are counting down from max_steps */
	  filter.Add_Complex( (*pt_complex_B) , max_steps - i);

	  /* We are done using pt_complex_A so we free its memory */
      delete pt_complex_A;
      /* Note that complex_A whas complex B in the previous iteration of the for loop */
      
      /* We now set pt_complex_A to point to complex_B */
      pt_complex_A = pt_complex_B;
    }
    else{
      /* If we were not successful, then we continue on, after managing our memory */
      delete pt_complex_B;
	  /* If the first threshold did not get verified, then we need to exit this subroutine. */
	  if (i == 0){
		  std::cout << "Fatal Error!! \nChoose a different";
		  if (sublevel){
			  std::cout << " bottom ";
		  }
		  else{
			  std::cout << " top ";
		  }
		  std::cout << "threshold. \nExiting Verified_Filtration ...\n";
		  break;
	  }
    }
  }
  /* We delete the memory allocated to pt_complex_A */ 
  delete pt_complex_A;
  // After we have created a verified filtration, we write to the file "FilteredComplex" all of the cell and boundary information of our CW filtration.
  std::cout << "\nWriting filtered CW complex to output file \n";
  filter.Export();
}

void CApproximation::Modify_Output(double bottom_threshold, double step_size, unsigned int max_steps, bool sublevel){
	std::cout << "Converting integer thresholds in [output_*.txt] files, \n";
	std::cout << "      to real thresholds in [real_output_*.txt] files.\n";
	
	/* We gather all the the nontrivial output files. */
	std::ifstream Betti_0("output_0.txt");
	std::ifstream Betti_1("output_1.txt");
	std::ifstream Betti_numbers("output_betti.txt");
	/* 
	We do not process output_2 ; 
	Since the sub/super-level sets are a subset of the plane, they have trivial second homology.
	*/

	/* We initialize a vector of pointers to input files. */
	std::vector< std::ifstream * > ptr_in_files;
	///* We add the addresses of both Betti_0 and Betti_1 to the vector. */
	ptr_in_files.push_back(&Betti_0);
	ptr_in_files.push_back(&Betti_1);


	
	/*We initialize output streams */
	std::ofstream mod_Betti_0("real_output_0.txt");
	std::ofstream mod_Betti_1("real_output_1.txt");
	std::ofstream mod_Betti_numbers("real_output_betti.txt");
	/* We initialize a vector of pointers to output files. */
	std::vector< std::ofstream * > ptr_out_files;
	///* We add the addresses of both Betti_0 and Betti_1 to the vector. */
	ptr_out_files.push_back(& mod_Betti_0);
	ptr_out_files.push_back(& mod_Betti_1);


	/* 
	Both Betti_0 and Betti_1 have only two numbers per line: the birth time and the death time.
	We initialize two integers to store these numbers
	*/
	int birth, death;
	double real_birth, real_death;

	/* We iterate through the files Betti_0 and Betti_1 */
	for (int i = 0; i < ptr_in_files.size(); i++){
		/* We modify the i^th betti numbers*/
		while ((*ptr_in_files[i]) >> birth >> death){
			/* We store the birth-death pair for a single persistence interval, and then convert it to their real values */
			real_birth = Convert_Threshold(birth, bottom_threshold, step_size, max_steps, sublevel);
			real_death = Convert_Threshold(death, bottom_threshold, step_size, max_steps, sublevel);
			/* We write the real values to an output file. */
			(*ptr_out_files[i]) << real_birth << " " << real_death << "\n";
		}
	}
	
	/* Now we modify the output_betti.txt file.	This file has four numbers on each line:
	The threshold
	The 0th betti number 
	The 1th betti number 
	The 2th betti number

	We only need to modify the first number.
	*/
	int integer_threshold;
	double real_threshold;
	int zero, one, two;
	/* We iterate through each line of the file. */
	while (Betti_numbers  >> integer_threshold >> zero >> one >> two){
		/* We convert the integer threshold to its real value */
		real_threshold = Convert_Threshold(integer_threshold, bottom_threshold, step_size, max_steps, sublevel);
		/* We write the real values to an output file. */
		mod_Betti_numbers << real_threshold << " " << zero << " " << one << " " << two << "\n";
	}

	

}

/*
***************************
******   PRIVATE    *******
***************************
*/

/* 
This function converts the integer threshold output from perseus to the correct real-valued threshold.
*/
double CApproximation::Convert_Threshold(int integer_threshold, double bottom_threshold, double step_size, unsigned int max_steps, bool sublevel){
	
	/* We initialize the output variable */
	double real_threshold;
	if (sublevel){
		/* We define the real-threshold for sub-level sets. */
		real_threshold = bottom_threshold + (integer_threshold - 1) * step_size;
	}
	else{
		/* We calculate what the top threshold is */
		double top_threshold = bottom_threshold + step_size * (max_steps - 1);
		/* We define the real-threshold for super-level sets. */
		real_threshold = top_threshold - (integer_threshold - 1) * step_size;
	}

	return real_threshold;
}
